//Illinois Open Source License
//
//University of Illinois
//Open Source License
//
//Copyright © 2008,    Board of Trustees of the University of Illinois.  All rights reserved.
//
//Developed by:
//
//MONET Research Group in the Department of Computer Science
//The University of Illinois at Urbana-Champaign
//http://cairo.cs.uiuc.edu

// Distributed Protocols Research Group in the Department of Computer Science
// The University of Illinois at Urbana-Champaign
// http://dprg.cs.uiuc.edu/

//Long Vu, longvu2@cs.uiuc.edu
//Indranil Gupta, indy@cs.uiuc.edu
//Klara Nahrstedt, klara@cs.uiuc.edu
//Jin Liang, jinliangcu@gmail.com
//
//Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal with the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimers.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimers in the documentation and/or other materials provided with the distribution.
//    * Neither the names of The Monet Group or The University of Illinois at Urbana-Champaign, nor the names of its contributors may be used to endorse or promote products derived from this Software without specific prior written permission.
//
//THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
//PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN 
//AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.

 
  
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <pthread.h>
#include <time.h>

#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include "peerElem.h"

#define SERVERPORT 8000
#define MAXBUF 1024
#define STARTOFFSET 37
#define STEP 16
#define DATAPACKET 150 // if packet size is more than 150 => this is (list of peers) returned packet

#define SERVER_TIMEOUT 5 // seconds
#define PEER_TIMEOUT 20
#define INIT_ROUND 500 // to send to the severs to initialize the list
#define INIT_REQUEST_RATE 5000 
#define CONSUMER_REQUEST_RATE 50000 
#define PEER_RECEIVED_NUM 1000 // this is to check when to stop
#define NEW_PEER 15 // this is to check when to stop

using namespace std;

//global declaration
char gs_sMess[MAXBUF];
char gs_pMess1[MAXBUF];
char gs_pMess2[MAXBUF];
char gs_outFile[MAXBUF];
char gs_serverListFile[MAXBUF];
unsigned short int gi_maxResponse; // number of max response we collect from a peer

vector<string> gs_serverList;
vector<CollectedPeer> g_peerList; // all collected peers
vector<ResponsivePeer> g_responsiveList; // all responsive peers

int gi_sockfd;
socklen_t gi_addrLen = sizeof(struct sockaddr);

int gi_newPeerNum = 0; 
int gi_peerReceivedNum = 0;
int gi_exitFlag = 0;

pthread_mutex_t g_mutexNewPeerNum;
pthread_mutex_t g_mutexReceivedNum;
pthread_mutex_t g_mutexPeerList;

//subfunction declaration
void init();
void convertHex(char* mess, char* retMess);
int getDecimalVal(char ch);
int peerExistCollectedList(unsigned long ip, unsigned short int port);
void writeSnapshot();
void writePartnerList();
void* producer(void* null);
void* consumer(void* null);
//---------------------------------------------------------------------------------
int main(int argc, char *argv[])
{

	cout << "number of parameters and testing: " << argc << endl; 
	if (argc != 7){
		cout << "Usage: crawler <severMess> <peerMess1> <peerMess2> <serverListFile> <maxResponse> <outFile>" << endl;
		exit(0);
	}

	pthread_attr_t attr;
	pthread_t pro, con;
	int i_status, i_rc;

	strcpy(gs_sMess,argv[1]);
	strcpy(gs_pMess1,argv[2]);
	strcpy(gs_pMess2,argv[3]);
	strcpy(gs_serverListFile,argv[4]);
	strcpy(gs_outFile,argv[6]);

	gi_maxResponse = strtol(argv[5], NULL, 10);

	g_peerList.clear();
	g_responsiveList.clear();

	gi_exitFlag = 0;
	pthread_mutex_init(&g_mutexPeerList, NULL);
	
	init();
	
	// After initial phase, we have a list of peers
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

	i_rc = pthread_create(&pro, &attr, producer, (void **)&i_status);
	if (i_rc){
         printf("ERROR; return code from pthread_create() producer is %d\n", i_rc);
    }
	
	i_rc = pthread_create(&con, &attr, consumer, (void **)&i_status);
	if (i_rc){
         printf("ERROR; return code from pthread_create() consumer is %d\n", i_rc);
    }
	
	pthread_join(pro, NULL);
	pthread_join(con, NULL);

	pthread_attr_destroy(&attr);
	pthread_mutex_destroy(&g_mutexPeerList);

	close(gi_sockfd);
	writeSnapshot();
	writePartnerList();
	cout << "Write to file and exit crawler ..." << endl;
	return 0;
}
//-----------------------------------------------------------------
void* producer(void* null){
	// listens on given socket and receive any incoming packets
	fd_set readfds;
	struct sockaddr_in tmp_addr;
	char buf[MAXBUF];
	int numbytes;
    unsigned long ipAddr;
    unsigned short int port;
	struct timeval tv;

	cout << "Entering producer of the crawler..." << endl;

	while (1){
		tv.tv_sec = PEER_TIMEOUT; 
		tv.tv_usec = 0;
		FD_ZERO(&readfds);
		FD_SET(gi_sockfd, &readfds);
		select(gi_sockfd+1, &readfds, NULL, NULL, &tv);
		if (FD_ISSET(gi_sockfd, &readfds)){
			memset(buf,'\0',MAXBUF);
			if ((numbytes=recvfrom(gi_sockfd,buf, MAXBUF, 0,
							   (struct sockaddr *)&tmp_addr, &gi_addrLen)) == -1) {
				perror("recvfrom server has error ");
				exit(1);
			}

			if (numbytes > DATAPACKET) {
				
				// update the flag of the responder
				unsigned long ipTmp = tmp_addr.sin_addr.s_addr;
				unsigned short int portTmp = ntohs(tmp_addr.sin_port);
				
				cout << "producer receive response from " << ipTmp<< endl;
				for (unsigned k = 0 ; k < g_peerList.size() ; k++){
					if ((g_peerList[k].ui_port == portTmp) && (g_peerList[k].ul_ipAddr == ipTmp)){
						pthread_mutex_lock(&g_mutexPeerList);
						g_peerList[k].setFlag(2); // this means the tmp_addr.sin_addr.s_addr responses
						pthread_mutex_unlock(&g_mutexPeerList);
						cout << "producer checks " << ipTmp << " and exists"<< endl;
						break;
					}
				}
				
				int indexResponder = -1;
				//cout << "producer checks responsive list "<< endl;
				for (unsigned k = 0 ; k < g_responsiveList.size() ; k++){
					
					if ((g_responsiveList[k].ui_port == portTmp) && (g_responsiveList[k].ul_ipAddr == ipTmp)){
						indexResponder = k;
						break;
					}
				}

				if (indexResponder < 0) // new responder
				{	
					ResponsivePeer peer(ipTmp, portTmp, 0);
					g_responsiveList.push_back(peer);
					indexResponder = g_responsiveList.size() - 1; // now the new element is at the end of the responsive list
				}				
				
				//cout << "producer add the responsive peer to the list at "<< indexResponder << ", responsive list size = " << g_responsiveList.size() << endl;
				
				// extract peers
				for (int i=STARTOFFSET + 12 ; i< numbytes ;i += STEP ){
					gi_peerReceivedNum ++;

					ipAddr 	= *((unsigned long*)(buf + i));
					port 	= *((unsigned short int *)(buf + i + 6));
					 if (peerExistCollectedList(ipAddr,port) < 0){
						 CollectedPeer peer(ipAddr, port, 1); // 1 means initial value of resposeFlag
						 pthread_mutex_lock(&g_mutexPeerList);
						 g_peerList.push_back(peer);
						 pthread_mutex_unlock(&g_mutexPeerList);
						 gi_newPeerNum ++;
					 }
					// cout << "producer receive add partners to " << g_responsiveList[indexResponder].ul_ipAddr << endl;
					 // add into the partner list of the responsive peer
					 g_responsiveList[indexResponder].addPartner(ipAddr, port, gi_maxResponse);
				}
				g_responsiveList[indexResponder].incResponseTime();
				cout << "Producer: Receive packets from " << inet_ntoa(tmp_addr.sin_addr) << " port = " << portTmp ;

				if (gi_peerReceivedNum >= PEER_RECEIVED_NUM){
					cout << "\nCHECK STOP: #peer received = "<< gi_peerReceivedNum << ", #new peers = "<< gi_newPeerNum << "\n";
					if (gi_newPeerNum >= NEW_PEER){
						gi_newPeerNum = 0;
						gi_peerReceivedNum = 0;
					}else{ // exit
						break;
					}
				}
			}
		} // end of if FD
		else{ // Time out
			break;
		}
	}// end of While
	gi_exitFlag = 1;
	cout << "Exit producer ..." << endl;
}
//-----------------------------------------------------------------
void* consumer(void* null){
	// use round robin approach
	cout << "Entering consumer ..." << endl;
	int numbytes;
	unsigned i = 0;
	struct sockaddr_in mPeer_addr;
	char convertedMess1[MAXBUF];
	char convertedMess2[MAXBUF];

	convertHex(gs_pMess1, convertedMess1);
	convertHex(gs_pMess2, convertedMess2);
	
	mPeer_addr.sin_family = AF_INET;         // host byte order
	
	while (i < g_peerList.size()){
			// prepare to probe peer
		    unsigned long ipAddr = g_peerList[i].ul_ipAddr;
		    unsigned short int port = g_peerList[i].ui_port;
			//if ((port > 0) && (g_peerList[i].ui_flag > 0)){
			if (port > 0){
				mPeer_addr.sin_port = htons(port);     // short, network byte order
				mPeer_addr.sin_addr.s_addr = ipAddr;
				cout << "Consumer sends to " << ipAddr << " port " << port << " index = "<< i << endl;
				memset(&(mPeer_addr.sin_zero), '\0', 8); // zero the rest of the struct
				if ((numbytes=sendto(gi_sockfd, convertedMess1, strlen(gs_pMess1), 0,
				 (struct sockaddr *)&mPeer_addr, sizeof(struct sockaddr))) == -1) {
					cout << "ERROR: Inside consumer: sendto " << inet_ntoa(mPeer_addr.sin_addr) << " port = " << g_peerList[i].ui_port << " has error " << endl;
					pthread_mutex_lock(&g_mutexPeerList);
					g_peerList[i].ui_flag = 0;
					pthread_mutex_unlock(&g_mutexPeerList);
				}else if ((numbytes=sendto(gi_sockfd, convertedMess2, strlen(gs_pMess2), 0,
				 	(struct sockaddr *)&mPeer_addr, sizeof(struct sockaddr))) == -1) {
					cout << "ERROR: Inside consumer: sendto " << inet_ntoa(mPeer_addr.sin_addr) << " port = " << g_peerList[i].ui_port << " has error " << endl;
					pthread_mutex_lock(&g_mutexPeerList);
					g_peerList[i].ui_flag = 0;
					pthread_mutex_unlock(&g_mutexPeerList);
				 }else{
					usleep(CONSUMER_REQUEST_RATE); 
				 } // end if sendto
			} // end if port > 0
			
			i++;
			if (i == g_peerList.size()){ // round robin
				i = 0;
			}
			if (gi_exitFlag){
				break;
			}
	}// end of While
	cout << "Exit consumer ..." << endl;
}
//-----------------------------------------------------------------
void init(){
	char convertedMess[MAXBUF];
	char buf[MAXBUF];
	int i, count, numbytes, len, probes;
    unsigned long ipAddr;
    unsigned short int port;
	struct sockaddr_in tmp_addr, mServer_addr;
	struct timeval tv;
	
	fd_set readfds;
	// get server.lst
	string s_serverName;
	ifstream serverFile (gs_serverListFile);
	if (serverFile.is_open())
	{
	    while (! serverFile.eof() )
	    {
	      getline (serverFile,s_serverName);
		  gs_serverList.push_back(s_serverName);
	    }
	    serverFile.close();
	}
	
	if ((gi_sockfd = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
		perror("socket creation has error ");
		exit(1);
	}
	
	convertHex(gs_sMess, convertedMess);
	len = strlen(gs_sMess);

	mServer_addr.sin_family = AF_INET;         // host byte order
	mServer_addr.sin_port = htons(SERVERPORT);     // short, network byte order

	count = 0;probes = 1;
	while (1){
		if (probes < INIT_ROUND){
			for(int index = 0; (unsigned)index < gs_serverList.size(); index++ ){
				inet_aton(gs_serverList[index].c_str(), &(mServer_addr.sin_addr));
				memset(&(mServer_addr.sin_zero), '\0', 8); // zero the rest of the struct
				if ((numbytes=sendto(gi_sockfd, convertedMess, len, 0,
				 (struct sockaddr *)&mServer_addr, sizeof(struct sockaddr))) == -1) {
					perror("sendto");
					exit(1);
				}
				usleep(INIT_REQUEST_RATE);
			} // end of FOR 
			probes++;
		}// end if INIT_ROUND
		
		tv.tv_sec = SERVER_TIMEOUT; 
		tv.tv_usec = 0;
		FD_ZERO(&readfds);
		FD_SET(gi_sockfd, &readfds);
		select(gi_sockfd+1, &readfds, NULL, NULL, &tv);
		if (FD_ISSET(gi_sockfd, &readfds)){
			memset(buf,'\0',MAXBUF);
			if ((numbytes=recvfrom(gi_sockfd,buf, MAXBUF, 0,
							   (struct sockaddr *)&tmp_addr, &gi_addrLen)) == -1) {
				perror("recvfrom server has error ");
				exit(1);
			}
			if (numbytes > DATAPACKET) {
				for (i=STARTOFFSET; i< numbytes ;i += STEP ){
					ipAddr = *((unsigned long*)(buf + i));
					port = *((unsigned short int *)(buf + i + 6));
					 if (peerExistCollectedList(ipAddr,port) < 0){
						 count ++;
						 CollectedPeer peer(ipAddr, port, 1); // initial value
						 g_peerList.push_back(peer); // we add all peers to the Collected Peer list
					 }
				}
			cout << "Packets from " << inet_ntoa(tmp_addr.sin_addr) << " port " << ntohs(tmp_addr.sin_port) << ", after " << probes << " times, have " << count << endl;
			}
		}else{ // Timeout
			break;
		}
	}
	cout << "Finish initial phase " << endl;
}
//========================================================
void writeSnapshot(){
	struct sockaddr_in ipCollectedPeer;
	ofstream myfile;
	myfile.open (gs_outFile);
	
	cout << "Write SNAPSHOT to file: " << gs_outFile << endl;

	for(int index = 0; (unsigned)index < g_peerList.size(); index++ ){
		ipCollectedPeer.sin_addr.s_addr = g_peerList[index].ul_ipAddr;
		//if (g_peerList[index].ui_flag > 1){
			myfile << inet_ntoa(ipCollectedPeer.sin_addr);
			myfile << "_";
			myfile << g_peerList[index].ui_port;
			myfile << "\n";
		//}
	}
	cout << "SNAPSHOT size: " <<  g_peerList.size() << endl;
	myfile.close();
}
//========================================================
void writePartnerList(){
	struct sockaddr_in ipResponder;
	struct sockaddr_in ipPartner;
	ofstream listfile;
	string listFileName = gs_outFile;
	listFileName.append("_responsive");
	listfile.open (listFileName.c_str());
	listfile << "Total responder = " << g_responsiveList.size() << endl;
	for(int index = 0; (unsigned)index < g_responsiveList.size(); index++ ){
		ipResponder.sin_addr.s_addr = g_responsiveList[index].ul_ipAddr;
		listfile << "<responder=";
		listfile << inet_ntoa(ipResponder.sin_addr);
		listfile << "_";
		listfile << g_peerList[index].ui_port;
		listfile << ">";
		int count = 0;
		for(int j = 0; (unsigned)j < g_responsiveList[index].partnerList.size(); j++ ){
			ipPartner.sin_addr.s_addr = g_responsiveList[index].partnerList[j].ul_ipAddr;
			listfile << inet_ntoa(ipPartner.sin_addr) << "_" << g_responsiveList[index].partnerList[j].ui_port<< " ";
			count ++;
		}
		listfile << "</responder, count=" << count << ">\n";
	}
	listfile.close();
	cout << "Number of responsive peers " << g_responsiveList.size() << endl;
}
//========================================================
void convertHex(char* mess, char* retMess){
int i,j, d0, d1;
j = 0;
	for (i = 0; (unsigned)i < strlen(mess) ; i = i+2 )
	{
		d0 = getDecimalVal(mess[i]);
		d1 = getDecimalVal(mess[i+1]);
		d0 = d0 * 16 + d1;
		retMess[j++] = d0;
	}
	retMess[j] = '\0';
}
//------------------------------------------------------------------------------------------
int getDecimalVal(char ch){
	switch (ch)	{
		case 'a': return 10;
		case 'b': return 11;
		case 'c': return 12;
		case 'd': return 13;
		case 'e': return 14;
		case 'f': return 15;
		default:
			return ch-48;
	}
}
//--------------------------------------------------
int peerExistCollectedList(unsigned long ip, unsigned short int port){
   for(int index = 0; (unsigned)index < g_peerList.size(); index++ )
   {
       if ((g_peerList[index].ul_ipAddr == ip) && (g_peerList[index].ui_port == port)) return index;
   }	   
   return -1;
}
//--------------------------------------------------
