//Illinois Open Source License
//
//University of Illinois
//Open Source License
//
//Copyright © 2008,    Board of Trustees of the University of Illinois.  All rights reserved.
//
//Developed by:
//
//MONET Research Group in the Department of Computer Science
//The University of Illinois at Urbana-Champaign
//http://cairo.cs.uiuc.edu

// Distributed Protocols Research Group in the Department of Computer Science
// The University of Illinois at Urbana-Champaign
// http://dprg.cs.uiuc.edu/

//Long Vu, longvu2@cs.uiuc.edu
//Indranil Gupta, indy@cs.uiuc.edu
//Klara Nahrstedt, klara@cs.uiuc.edu
//Jin Liang, jinliangcu@gmail.com
//
//Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal with the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimers.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimers in the documentation and/or other materials provided with the distribution.
//    * Neither the names of The Monet Group or The University of Illinois at Urbana-Champaign, nor the names of its contributors may be used to endorse or promote products derived from this Software without specific prior written permission.
//
//THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
//PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN 
//AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.

#ifndef COLLECTEDPEER_H_
#define COLLECTEDPEER_H_

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <pthread.h>
#include <time.h>
#include <iostream>
#include <fstream>
#include <vector>


class CollectedPeer{
		public:
		unsigned long ul_ipAddr;
		unsigned short int ui_port;
		unsigned ui_flag;
		CollectedPeer(unsigned long ul_ipAddr_tmp, unsigned short int ui_port_tmp, unsigned ui_flag_tmp){
			ul_ipAddr 		= ul_ipAddr_tmp; 
			ui_port 		= ui_port_tmp;
			ui_flag			= ui_flag_tmp;
		}
		
		void setFlag(unsigned ui_flag_tmp){
			ui_flag			= ui_flag_tmp;
		}
		unsigned getFlag(){
			return ui_flag;
		}

};

//-------------------------------------------------------------------------
class ResponsivePeer{
	public:
		unsigned long ul_ipAddr;
		unsigned short int ui_port;
		unsigned short int ui_responseTime;
		std::vector<CollectedPeer> partnerList;
		
		ResponsivePeer(unsigned long ul_ipAddr_tmp, unsigned short int ui_port_tmp, unsigned short int ui_responseTmp){
			ul_ipAddr 		= ul_ipAddr_tmp; 
			ui_port 		= ui_port_tmp;
			ui_responseTime	= ui_responseTmp;
			
			partnerList.clear(); // initialize by empty list
		}

		inline void incResponseTime(){
			ui_responseTime ++;
		}
		
		void addPartner(unsigned long ul_ipAddr_tmp, unsigned short int ui_port_tmp, unsigned short int maxResponse);
		bool peerExist(unsigned long ip, unsigned short int port);
			
};

bool ResponsivePeer::peerExist(unsigned long ul_ipAddr_tmp, unsigned short int ui_port_tmp){
		for(int index = 0; (unsigned)index < partnerList.size(); index++ )
		{
			if ((partnerList[index].ul_ipAddr == ul_ipAddr_tmp) && (partnerList[index].ui_port == ui_port_tmp)) return true;
		}	   
		return false;
}

void ResponsivePeer::addPartner(unsigned long ul_ipAddr_tmp, unsigned short int ui_port_tmp, unsigned short int maxResponse){
	if ((ui_responseTime < maxResponse) && (! peerExist(ul_ipAddr_tmp, ui_port_tmp)) ){
		CollectedPeer peer(ul_ipAddr_tmp,ui_port_tmp,1); // by default the flag = 1
		partnerList.push_back(peer);
	}
}
		
#endif /*COLLECTEDPEER_H_*/
